# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import json

from luisy.code_inspection import create_hashes
from luisy.testing import get_all_dependencies
from luisy import (
    Task,
    ExternalTask,
)
from luisy import Config
from luisy.targets import ParquetDirTarget
from luisy.config import (
    add_working_dir,
    remove_working_dir,
    change_working_dir,
)

logger = logging.getLogger(__name__)


class TaskHashEntry:
    """
    This class manages all available hash entries in a luisy run. Each task of luisy has its own
    hash depending on the code version. So in one luisy run, a task can have multiple hashes:
    - hash in local `luisy.hash` file
    - hash in cloud `luisy_hash` file (cloud)
    - hash that was computed in this current run

    Most of the methods of this class do comparisons between those hashes to give luisy an
    idea of what needs to be executed, downloaded, uploaded and which hashes need to be refreshed
    in `.luisy.hash` file online or locally.

    Args:
        hash_new (str): hash of current luisy run
        hash_local (str or None): hash of local 'luisy.hash' file (if available)
        hash_cloud (str or None): hash of cloud 'luisy.hash' file (if available)
        task (luisy.Task): instance of task which all the hashes belong to
        filename (str): output filename of task (key inside `luisy.hash` file)

    Note:
        Can this be moved inside a Task instance?

    """

    def __init__(self, hash_new, hash_local=None, hash_cloud=None, task=None, filename=None):
        self.hash_new = hash_new
        self.hash_local = hash_local
        self.hash_cloud = hash_cloud
        self.task = task
        self.filename = filename

    def in_cloud(self):
        """
        Returns:
            bool: if Task has a hash saved in the cloud
        """
        return self.hash_cloud is not None

    def has_outdated_cloud_hash(self):
        """
        Returns:
            bool: if Task has a hash saved in the cloud and this is outdated
        """
        return self.hash_cloud != self.hash_new and self.in_cloud()

    def in_local(self):
        return self.hash_local is not None

    def is_downloadable(self):
        return self.in_cloud() and not self.has_outdated_cloud_hash()

    def has_outdated_local_hash(self):
        return self.hash_local != self.hash_new and self.in_local()

    def is_external_task(self):
        return isinstance(self.task, ExternalTask)

    def has_parquet_output(self):
        return issubclass(self.task.target_cls, ParquetDirTarget)

    def needs_local_execution(self):
        return (self.has_outdated_local_hash() or not self.in_local()) and not \
            self.is_downloadable()

    def needs_to_be_downloaded(self):
        return (self.has_outdated_local_hash() or not self.in_local()) and self.is_downloadable()


class HashSynchronizer(object):
    """
    Class to keep hashmappings of a luisy run synched. Calling the HashSynchronizer.initialize()
    loads the following Hashmappings, that will be managed in this class:
    - current run of luisy
    - local `.luisy.hash`
    - cloud `.luisy.hash`

    This class manages the synchronization between those hashmappings when luisy did some updates
    on them. It also offers some functions that gives the user information on which tasks to
    download/upload/rerun/...

    Args:
        tasks(dict): values are Tasks of current luisy run, keys are corresponding filenames.
        root_task(luisy.Task): Root task of the current luisy run (needed to get `project name`)

    """

    def __init__(self, tasks, root_task):
        self.tasks = tasks
        self.root_task = root_task
        self.hash_entries = []

    def check_params(self):
        assert self.root_task in self.tasks.values()
        task_filenames = [remove_working_dir(task.output().path) for task in self.tasks.values()]

        assert all([filename in task_filenames for filename in self.hashmapping_new.hashes.keys()])

    def initialize(self):
        """
        Initializes all hashmappings needed in this class:
        - hash_mapping_new: computed by current pipeline
        - hash_mapping_local: loaded from local `luisy.hash` file
        - hash_mapping_cloud: loaded from `luisy.hash` file in cloud

        Also sets up list of hash_entries, which are used to compare hashmappings.

        """
        self._initialize_hashmappings()
        hashes_new = self.hashmapping_new.hashes
        hashes_local = self.hashmapping_local.hashes
        hashes_cloud = self.hashmapping_cloud.hashes
        self.check_params()
        for filename in hashes_new.keys():
            self.hash_entries.append(TaskHashEntry(
                hash_new=hashes_new[filename] if filename in hashes_new.keys() else None,
                hash_local=hashes_local[filename] if filename in hashes_local.keys() else None,
                hash_cloud=hashes_cloud[filename] if filename in hashes_cloud.keys() else None,
                task=self.tasks[filename] if filename in self.tasks.keys() else None,
                filename=filename,
            ))

    def _initialize_hashmappings(self):
        self.hashmapping_new = self.compute_new_hashes()
        self.hashmapping_local = self.read_local_hash_file()
        self.hashmapping_cloud = self.read_cloud_hash_file()
        self._add_external_cloud_hashes()

    def compute_new_hashes(self):
        logger.info('Compute hashes of tasks')
        return HashMapping.compute_from_tasks(
            tasks=self.tasks,
            project_name=self.root_task.project_name
        )

    def read_local_hash_file(self):
        logger.info('Read Hashes from disk')
        hashes = HashMapping.load_from_local(
            project_name=self.root_task.project_name
        )
        hashes.clean_up()
        return hashes

    def read_cloud_hash_file(self):
        if Config().get_param('download') or Config().get_param('upload'):
            logger.info('Read Hashes from cloud')
            hashes = HashMapping.load_from_cloud(
                project_name=self.root_task.project_name
            )
            hashes.clean_up()
            return hashes
        else:
            return HashMapping(hashes={}, local=False, project_name=self.root_task.project_name)

    def _add_external_cloud_hashes(self):
        """

        It might be that some files in the cloud are added by a none-luisy system, like a Databricks
        cluster that writes some parquets to the blob-storage correctly. Those files cannot be
        versioned with luisy and we have to assume that these files are immutable. Thus, we will
        check if any tasks in the pipeline point to such files and will add their hashes.
        """
        for filename, task in self.tasks.items():
            if issubclass(task.target_cls, ParquetDirTarget):

                if self.hashmapping_cloud.get_hash(filename) is not None:
                    continue

                self.hashmapping_cloud.update(
                    filepath=filename,
                    task_hash=self.hashmapping_cloud.get_hash(filename)
                )

    def get_outdated_local_hash_entries(self):
        """
        Out dated tasks without external tasks as external tasks should not be reran at all

        Returns:
            list: hash entries of tasks that are outdated (no external tasks included)

        """
        return [
            hash_entry for hash_entry in self.hash_entries if
            hash_entry.has_outdated_local_hash() and not hash_entry.is_external_task()
        ]

    def get_new_local_hash_entries(self):
        """
        Tasks that are non existent on local machine

        Returns:
            list: hash entries that are not present in local luisy.hash

        """
        return [hash_entry for hash_entry in self.hash_entries if not hash_entry.in_local()]

    def get_downloadable_hash_entries(self):
        """
        Identifies which of the files available in the cloud have the hash needed for the local
        execution.

        Returns:
            list: List of hash entries that can be downloaded from the cloud having the latest hash

        """
        return [hash_entry for hash_entry in self.hash_entries
                if hash_entry.is_downloadable()]

    def get_new_cloud_hash_entries(self):
        """
        Non existent tasks in the cloud

        Returns:
            list: hash entries that do not exist in cloud yet

        """
        return [hash_entry for hash_entry in self.hash_entries
                if not hash_entry.in_cloud() and not hash_entry.has_parquet_output()]

    def get_outdated_cloud_hash_entries(self):
        """
        Outdated tasks in cloud

        Returns:
            list: hash entries of outdated tasks in cloud

        """
        return [hash_entry for hash_entry in self.hash_entries
                if hash_entry.has_outdated_cloud_hash() and not hash_entry.has_parquet_output()]

    def get_all_hash_entries_that_need_a_run(self):
        """
        Outdated or non existing local tasks

        Returns:
            list: Hash entries of tasks that will be refreshed in this luisy run

        """
        return list(set(self.get_new_local_hash_entries() + self.get_outdated_local_hash_entries()))

    def get_local_execution_hash_entries(self):
        """
        Tasks that need to be executed locally because they cannot be downloaded

        Returns:
            list: hashmappings with tasks that need to be executed locally

        """
        return [hash_entry for hash_entry in self.hash_entries
                if (hash_entry.needs_local_execution())]

    def get_download_hash_entries(self):
        """
       Non existent local tasks that are available with their newest versions in the cloud

        Returns:
            list: hashmappings that will be downloaded

        """
        return [hash_entry for hash_entry in self.hash_entries
                if (hash_entry.needs_to_be_downloaded())]

    def get_upload_hash_entries(self):
        """
        All outdated or non existing tasks in cloud

        Returns:
            list: hash entries to upload

        """
        return self.get_outdated_cloud_hash_entries() + self.get_new_cloud_hash_entries()

    def clear_outdated_tasks(self):
        """
        Clears the output of all tasks that are outdated locally

        Note:
            This can also be implemented for cloud clearing here. Currently the rerun tasks in
            the cloud are just overwritten when saved online

        """
        for hash_entry in self.get_outdated_local_hash_entries():
            hash_entry.task.clean()

    def synchronize_local_hashes(self, failed_tasks):
        """
        This function updates all local hashes to the hashes from the current run. Failed tasks
        can be given as an argument to prevent luisy from setting hashes that have not been
        executed.

        Args:
            failed_tasks (list): Failed tasks in luigi run

        """
        executed_tasks = self.get_new_local_hash_entries() + self.get_outdated_local_hash_entries()
        for hash_entry in executed_tasks:
            if hash_entry.task in failed_tasks:
                logger.warning(f"Task for {hash_entry.filename} failed, will not update hash")
                continue
            outfile = hash_entry.filename
            self.hashmapping_local.update(
                filepath=outfile,
                task_hash=self.hashmapping_new.get_hash(outfile)
            )
            self.hashmapping_local.save()

    def synchronize_cloud_hashes(self, failed_tasks, upload_tasks=True):
        """
        This function updates all online/cloud hashes to the hashes from the current run. Failed
        tasks can be given as an argument to prevent luisy from setting hashes that have not been
        uploaded. In addition, if `upload_tasks` is True, this method uploads all the tasks that
        have outdated hash entries as well.

        Args:
            failed_tasks (list): Failed tasks in luigi run

        """

        for hash_entry in self.get_upload_hash_entries():
            # Only upload successfully tasks
            if hash_entry.task in failed_tasks:
                logger.warning(f"Task for {hash_entry.filename} failed, will not upload")
                continue

            if upload_tasks:
                hash_entry.task.upload(overwrite=True)
            self.hashmapping_cloud.update(
                filepath=hash_entry.filename,
                task_hash=hash_entry.hash_new,
            )
            # Update the cloud file after each upload to not start all over again in terms
            # of crashes
            self.hashmapping_cloud.save()

    def set_files_to_download(self):
        """
        Sets files that need to be downloaded when luigi is executed. Files that need to be
        downloaded will be set inside the Config() singleton and checked in the
        :py:meth`luisy.Target.exists()` method.

        Note:
            Can this be done dynamically during luigi run?
        """
        Config()._files_to_download = [add_working_dir(hash_entry.filename) for hash_entry in
                                       self.get_download_hash_entries()]


class HashMapping(object):
    """
    Maps generic file identifiers to hashes of the tasks. To work in different working-dirs, the
    file identifiers are the paths to the file with the working-dir removed.

    Args:
        hashes (dict): Dict where the keys are file identifiers and the values are hashes of the
            tasks creating the respective files.
        local (bool): Whether the files of the hashes are located locally or remotely (i.e., the
            cloud).
        project_name (str): Identifier of the project. Typically the string marked by the
            decorator :py:func:`luisy.decorators.project_name`
    """

    def __init__(self, hashes, local=True, project_name=None):
        self.hashes = hashes
        self.local = local
        self.project_name = project_name

    @staticmethod
    def get_local_hash_path(project_name):
        return os.path.join(
            Config().working_dir,
            project_name,
            '.luisy.hash'
        )

    @staticmethod
    def get_cloud_hash_path(project_name):
        return change_working_dir(
            path=HashMapping.get_local_hash_path(project_name),
            dir_current=Config().working_dir,
            dir_new='')

    def get_hash(self, filepath):
        return self.hashes.get(filepath, None)

    @classmethod
    def load_from_local(cls, project_name):
        """
        Loads the mapping from the local :code:`.luisy.hash` file available in the working-dir of
        the project.

        Args:
            project_name (str): Name of the project.

        Returns:
            HashMapping: The mapping of the hashes

        """

        filepath = HashMapping.get_local_hash_path(project_name)

        logger.info(f'Read hashes from {filepath}')
        if os.path.exists(filepath):
            with open(filepath, 'r') as f:
                hashes = json.load(f)
        else:
            # If no hash-file exists, set up one
            logger.info('No hash-file there, will create one. May have to run some tasks')
            hashes = {}

        return HashMapping(hashes, local=True, project_name=project_name)

    def update(self, filepath=None, task_hash=None):
        self.hashes[filepath] = task_hash

    @classmethod
    def load_from_cloud(cls, project_name):
        """
        Loads the mapping from the cloud version of :code:`.luisy.hash` in the cloud-storage of the
        project.

        Args:
            project_name (str): Name of the project.

        Returns:
            HashMapping: The mapping of the hashes

        """

        fs = Config().fs

        filepath = HashMapping.get_cloud_hash_path(project_name)

        if fs.exists(filepath):
            logger.info('File available. Start download')
            hashes = fs.load_json(filepath)

        else:
            logger.info('No hash file in cloud')
            hashes = {}

        return HashMapping(hashes, local=False, project_name=project_name)

    @staticmethod
    def compute_from_tasks(tasks, project_name):
        """
        Computes the mapping from a dict of tasks.

        Args:
            tasks (dict): Dict where the values are :py:class:`luisy.Task` objects and the keys are
                their file identifiers (i.e., the filepaths without the working-directory).
            project_name (str): Name of the project.

        Returns:
            HashMapping: The mapping of the hashes
        """
        hashes = compute_hashes(
            tasks,
            requirements_path=Config().get_param('requirements_path')
        )
        return HashMapping(hashes, local=True, project_name=project_name)

    def save(self):
        """
        Saves the hashes either to disc (if :code:`local=True`) or to the cloud storage (if
        :code:`local=False`)
        """
        if self.local:
            filepath = HashMapping.get_local_hash_path(self.project_name)
            os.makedirs(filepath.replace(os.path.basename(filepath), "", ), exist_ok=True)
            logger.info(f'Write hashes to {filepath}')
            with open(filepath, 'w+') as f:
                json.dump(self.hashes, f)
        else:
            fs = Config().fs
            logger.info('Upload hashes to cloud')
            filepath = HashMapping.get_cloud_hash_path(self.project_name)
            fs.save_json(filepath, self.hashes)

    def items(self):
        return self.hashes.items()

    def exists(self, filepath):
        """
        Tests whether the output behind filepath exists, either locally or in the cloud.
        """

        if self.local:
            filepath_local = add_working_dir(filepath)
            return os.path.exists(filepath_local)
        else:
            fs = Config().fs
            return fs.exists(filepath)

    def clean_up(self):
        """
        Checks if for all existing hashes the outfiles are there. If not, the hashes are removed.
        """

        filepaths = list(self.hashes.keys())
        for filepath in filepaths:
            # TODO Better integration?
            if filepath.endswith('DeltaTable'):
                continue
            if not self.exists(filepath):
                logger.info(f"{filepath} not existing any more, clean up")
                self.hashes.pop(filepath)


def get_upstream_tasks(task):
    """
    Computes all upstream tasks of the given tasks by enumerating the computational DAG.

    Args:
        task (luisy.base.Task): A luisy task object

    Returns:
        dict: Dict where the values are the upstream tasks and the keys are the respective output
        files.
    """
    tasks = {}

    for dep in get_all_dependencies(task):
        if isinstance(dep, Task):
            tasks[remove_working_dir(dep.output().path)] = dep
        else:
            logger.warning(
                'None-luisy Tasks in your pipeline detected.'
                'Those tasks may prevent the rerun-functionality to work correctly!'
            )
    return tasks


def compute_hashes(tasks, requirements_path=None):
    """
    Computes the hashes for all tasks

    Args:

        tasks (dict): Dict where the values are :py:class:`luisy.Task` objects and the keys are the
            paths to their output.
        requirements_path (str): Path to the code:`requirement.txt`. Defaults to `None` and will
            be read-off the config.
    Returns:
        dict: Dict where the keys are the outfiles of the tasks handed in and the values are the
        hashes.

    TODO:
        Can this be moved to TaskHashEntry?

    """
    filenames = list(tasks.keys())
    task_list = [tasks[f] for f in filenames]

    tasks_objects = []
    tasks_classes = []

    for task in task_list:
        task_class = type(task)
        if task_class not in tasks_classes:
            tasks_classes.append(task_class)
            tasks_objects.append(task)

    hashes = create_hashes(tasks_classes, requirements_path=requirements_path)
    hash_dict = dict(zip(tasks_classes, hashes))

    return {filename: hash_dict[type(task)] for filename, task in tasks.items()}



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0
import pickle
import json
import luigi
import pandas as pd
import os
import logging
import re
from pyspark import SparkContext
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.errors.exceptions.base import AnalysisException
from luisy.config import (
    Config,
    change_working_dir,
)
from luisy.helpers import get_df_from_parquet_dir

logger = logging.getLogger(__name__)


class LuisyTarget(luigi.LocalTarget):
    """

    Args:
        path (str): Path to the file
        working_dir (str): The working dir
        download (bool): Whether the file should be downloaded from the cloud if not available
            locally
    """

    requires_pandas = False

    @property
    def fs(self):
        return Config().fs

    def exists(self):
        raise NotImplementedError()

    def write(self, obj):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()

    def remove(self):
        raise NotImplementedError()

    def _try_to_upload(self, overwrite=False):
        raise NotImplementedError()

    def _try_to_download(self):
        raise NotImplementedError()


class LocalTarget(LuisyTarget):
    file_ending = None
    requires_pandas = True

    def __init__(self, path, **kwargs):
        luigi.LocalTarget.__init__(
            self,
            path=path,
            format=None,
            is_tmp=False
        )
        self.kwargs = kwargs

    def is_folder(self):
        """
        Checks if the output is a folder.

        .. note::
            This may has to be improved in future as we (silently) assume that any file has file
            extension.

        """
        _, file_ending = os.path.splitext(os.path.basename(self.path))
        return file_ending == ''

    def make_dir(self, path):
        """
        Creates the local path `path`.

        """
        logger.info(f'Create dir {path}')
        os.makedirs(path, exist_ok=True)

    def remove(self):
        if os.path.exists(self.path):
            os.remove(self.path)
        else:
            logger.info(f'Target {self.path} not existing, nothing to remove')

    def exists(self):
        """
        Checks if the file exists. If :code:`download` is set to :code:`True`, then it is checked
        whether the file is available in the cloud if it is not available locally. If it is
        available, it is downloaded.
        """

        file_exists_locally = os.path.exists(self.path)

        if not file_exists_locally and Config().get_param('download'):
            if self.path in Config()._files_to_download:
                logger.info('Task not complete locally, check if file is available in cloud')
                self._try_to_download()

            # Check again if file exists
            file_exists_locally = os.path.exists(self.path)

        return file_exists_locally

    def _try_to_download(self):
        """
        Tries to download the target file from the cloud
        """
        path = change_working_dir(
            path=self.path,
            dir_current=Config().working_dir,
            dir_new='')

        if self.fs.exists(path):
            logger.info('File available. Start download')

            if self.is_folder():
                logger.info('Download folder')

                self.fs.download(
                    source=path,
                    dest=self.path)
            else:
                logger.info('Download file')
                self.fs._download_file(
                    source=path,
                    dest=self.path,
                )
        else:
            logger.info('File not in cloud.')

    def _try_to_upload(self, overwrite=False):
        """
        Tries to upload the target file to the cloud
        """
        path = change_working_dir(
            path=self.path,
            dir_current=Config().working_dir,
            dir_new='')

        if self.is_folder():
            raise NotImplementedError('Not implemented yet')

        logger.info('File not available. Start upload')
        self.fs._upload_file(
            source=self.path,
            dest=path,
            overwrite=overwrite,
        )


class CloudTarget(LuisyTarget):

    def __init__(self, path, **kwargs):
        LuisyTarget.__init__(
            self,
            path=path,
            format=None,
            is_tmp=False
        )
        self.kwargs = kwargs

    def _try_to_upload(self, overwrite=False):
        """No need to upload cloud targets."""
        pass

    def _try_to_download(self):
        """No need to download cloud targets."""
        pass


class SparkTarget(CloudTarget):
    """
    this abstract class is for targets working with spark instances.
    """
    @property
    def spark(self):
        try:
            return Config().spark
        except AttributeError:
            raise AttributeError(
                "spark session was not found in config. Make sure to have all the databricks "
                "parameters set in order to start the spark session!"
            )


class DeltaTableTarget(SparkTarget):
    # TODO: Can we get rid of fileending
    file_ending = 'DeltaTable'

    def __init__(
            self,
            outdir=None,
            schema="schema",
            catalog="catalog",
            table_name=None
    ):
        self.outdir = outdir
        self.table_name = table_name
        self.schema = schema
        self.catalog = catalog

    def make_dir(self, path):
        # TODO: Nothing to do here, adapt interface?
        pass

    def remove(self):
        self.spark.sql(f"DROP TABLE IF EXISTS {self.table_uri}")

    @property
    def table_uri(self):
        return f"{self.catalog}.{self.schema}.{self.table_name}"

    @property
    def path(self):
        # TODO: Path here is more an identifier that shows up in `.luisy.hashes`
        return os.path.join(
            self.outdir,
            f"{self.table_uri}.{self.file_ending}",
        )

    def exists(self):
        """
        Checks whether the Deltatable exists.

        Note:
            Ideally, we would call `self.spark.catalog.tableExists(self.table_uri)` to check whether
            the table exists, but this always returns `False`.
        """
        try:
            self.spark.sql(f"SELECT 1 from {self.table_uri}")
            return True
        except AnalysisException:
            return False

    def write(self, df: SparkDataFrame):
        """

        Args:
            df (pyspark.sql.DataFrame): Dataframe that should be written to delta table
        """
        logger.info(f"Write to {self.table_uri}")
        df.write.mode("overwrite").saveAsTable(self.table_uri)

    def read(self):
        return self.spark.table(self.table_uri)


class AzureBlobStorageTarget(SparkTarget):
    file_ending = ""

    def __init__(
            self,
            outdir=None,
            endpoint=None,
            directory=None,
            inferschema=False,
            file_format="parquet",
    ):
        self.outdir = outdir
        self.endpoint = endpoint
        self.directory = directory
        self.inferschema = inferschema
        self.file_format = file_format

    def make_dir(self, path):
        pass

    def remove(self):
        """
        we do not remove files from azure blob storage, but we always overwrite.
        """
        pass

    @property
    def blob_uri(self):
        return os.path.join(self.endpoint, self.directory)

    @property
    def path(self):
        # TODO: Path here is more an identifier that shows up in `.luisy.hashes`
        return os.path.join(
            self.outdir,
            self.endpoint,
            self.directory,
        )

    def exists(self):
        """
        Checks whether the file exists in Azure Blob Storage

        """
        try:
            self.spark.read.format(self.file_format).load(self.blob_uri).limit(1).count()
            return True
        except Exception:
            return False

    def write(self, df):
        """
        Write Pyspark DataFrame to Azure Blob Storage
        Args:
            df (pyspark.DataFrame): DataFrame that is to be stored in Azure Blob Storage
        """
        df.write.format(self.file_format).mode("overwrite").save(self.blob_uri)

    def read(self):
        """
        Read object from Azure Blob
        """

        return self.spark.read.format(self.file_format).load(
            self.blob_uri,
            inferschema=self.inferschema,
        )


class PickleTarget(LocalTarget):
    file_ending = 'pkl'

    def write(self, obj):
        with open(self.path, 'wb') as f:
            pickle.dump(obj, f)

    def read(self):
        with open(self.path, 'rb') as f:
            return pickle.load(f)


class FeatherTarget(LocalTarget):
    file_ending = 'feather'

    def write(self, df):
        df.to_feather(self.path)

    def read(self):
        return pd.read_feather(self.path)


class HDFTarget(LocalTarget):
    file_ending = 'hdf'

    def write(self, df):
        df.to_hdf(self.path, key='data')

    def read(self):
        return pd.read_hdf(self.path)


class XLSXTarget(LocalTarget):
    file_ending = 'xlsx'

    def write(self, df):
        df.to_excel(self.path)

    def read(self):
        return pd.read_excel(self.path, **self.kwargs)


class CSVTarget(LocalTarget):
    file_ending = 'csv'

    def write(self, df):
        df.to_csv(self.path, index=False, **self.kwargs)

    def read(self):
        return pd.read_csv(self.path, **self.kwargs)


class ParquetDirTarget(LocalTarget):
    file_ending = ''
    requires_pandas = False

    def _make_spark_df(self, df):
        if Config().spark is None:
            Config().set_param('spark', SparkContext())

        return Config().spark.createDataFrame(df)

    def write(self, df):
        if isinstance(df, pd.DataFrame):
            df = self._make_spark_df(df)
        df.write.parquet(self.path)

    def read(self):
        return get_df_from_parquet_dir(self.path)


class JSONTarget(LocalTarget):
    file_ending = 'json'
    requires_pandas = False

    def write(self, dct):
        assert isinstance(dct, dict)
        with open(self.path, 'w') as f:
            json.dump(dct, f, indent=4, separators=(',', ': '))

    def read(self):
        with open(self.path, 'r') as f:
            return json.load(f)


class DirectoryTarget(LocalTarget):
    """
    Target to read directories. The read_file method has to be filled by the user and passed to
    target by decorating it.
    """
    file_ending = ''
    regex_pattern = None

    def __init__(self, path, **kwargs):
        luigi.LocalTarget.__init__(
            self,
            path=path,
            format=None,
            is_tmp=False
        )
        if self.regex_pattern is None:
            self.regex_pattern = ''

        self.kwargs = kwargs

    def write(self):
        raise NotImplementedError("Coming soon...")

    def is_valid_file(self, filename):
        """
        Checks if file matches given regex

        Args:
            filename (str): file that need to be checked against regex

        Returns:
            bool: flag if file matches regex

        """
        return re.search(self.regex_pattern, filename)

    def read(self):
        for i, filename in enumerate(os.listdir(self.path)):
            if self.is_valid_file(filename):
                content = self.read_file(os.path.join(self.path, filename))
                if content is not None:
                    yield content

    def read_file(self, file):
        raise NotImplementedError('Needs to be passed by user. See '
                                  'decorators.make_directory_output')



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

from luisy.tasks.base import Task # noqa
from luisy.tasks.base import ExternalTask # noqa
from luisy.tasks.base import ConcatenationTask # noqa
from luisy.tasks.base import WrapperTask # noqa
from luisy.tasks.base import SparkTask # noqa



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import pandas as pd
import unittest
import luigi
import pickle
import copy
from luigi.tools.deps import get_task_requires
from contextlib import ExitStack

from luisy import Config
from luisy.config import get_default_params

from luigi.execution_summary import LuigiStatusCode

from pyspark.errors.exceptions.connect import AnalysisException

from unittest.mock import (
    patch,
    Mock,
)


def mock_target(task, output_store):

    target = Mock()

    def read():
        return copy.deepcopy(output_store[task])

    def exists():
        return task in output_store

    def write(obj):
        output_store[task] = obj

    target.read = read
    target.exists = exists
    target.write = write

    return patch.object(task, "output", return_value=target)


def debug_run(tasks, return_summary=True):
    """
    This is a helper Method which allows to run pipelines in tests and bypassing cli. This method
    returns some useful information from the build as well in form of a dict.

    Args:
        tasks(list or luigi.Task): List of Tasks that should be run or a single
            :py:class:`luigi.Task`

    Returns:
        dict: Dictionary with information about the executed tasks. I.e. dict['completed']
            contains the tasks that have been completed in this run or dict['already_done']
            contains tasks that have been completed before running luigi
    """
    if isinstance(tasks, luigi.Task):
        tasks = [tasks]
    result = luigi.build(tasks, detailed_summary=True, local_scheduler=True)

    if return_summary:
        return luigi.execution_summary._summary_dict(result.worker)
    else:
        return result


def get_all_dependencies(task):

    deps = get_task_requires(task)
    for task_ in deps:
        deps = deps.union(get_all_dependencies(task_))

    deps.add(task)
    return deps


def create_file(file, val):
    with open(file, 'wb') as f:
        pickle.dump(val, f)


class LuisyTestCase(unittest.TestCase):

    def get_execution_summary(self, task, existing_outputs=None):
        result, _ = self._run_pipeline(task=task, existing_outputs=existing_outputs)
        return luigi.execution_summary._summary_dict(result.worker)

    def _run_pipeline(self, task, existing_outputs=None):
        """
        Runs the given task after providing the existing outputs.
        """

        create_testing_config(working_dir='/some/working_dir/')

        if existing_outputs is None:
            existing_outputs = []

        output_store = {
            t: o for t, o in existing_outputs
        }

        all_tasks = get_all_dependencies(task)

        patches = [mock_target(t, output_store) for t in all_tasks]

        with ExitStack() as stack:
            for patch_ in patches:
                stack.enter_context(patch_)

            result = debug_run([task], return_summary=False)
            return result, output_store

    def run_pipeline(self, task, existing_outputs=None, return_store=False):
        """
        Method that can be used to run luigi pipelines. You are able to pass existing outputs
        which mock a specific task's output. This helps the user in testing without mocking his
        task output or creating a temporary directory.

        Args:
            task (luigi.Task): Task that should be run
            existing_outputs (list): List of Tuples, where first element of tuple is the
                Taskobject whose output should be mocked and the second element is the content of
                the mocked task
            return_store (bool): If true, the whole store gets returned as a dict. The store
                contains all the outputs after the execution of the pipeline, where the tasks are
                the keys and the values are their outputs. If false, only the output of the given
                task is returned.

        Returns:
            dict or object: depending on `return_store` value, a dict with output of all tasks or
                just the output of the root_task
        """
        result, output_store = self._run_pipeline(task=task, existing_outputs=existing_outputs)

        if not result.status == LuigiStatusCode.SUCCESS:
            result_dict = luigi.execution_summary._summary_dict(result.worker)
            self.fail(result_dict)

        if return_store:
            return output_store
        else:
            return output_store[task]

    def assertSuccess(self, task, existing_outputs=None):
        result, _ = self._run_pipeline(task=task, existing_outputs=existing_outputs)
        self.assertEqual(result.status, LuigiStatusCode.SUCCESS)

    def assertFail(self, task, existing_outputs=None):
        result, _ = self._run_pipeline(task=task, existing_outputs=existing_outputs)
        self.assertNotEqual(result.status, LuigiStatusCode.SUCCESS)

    def assertMissing(self, task, existing_outputs=None):
        result, _ = self._run_pipeline(task=task, existing_outputs=existing_outputs)
        self.assertEqual(result.status, LuigiStatusCode.MISSING_EXT)


class DFMock(Mock):

    def __init__(self, table_uri, data):
        Mock.__init__(self)

        self.table_uri = table_uri
        self.data = data

    @property
    def write(self):
        return self

    def mode(self, *args):
        return self

    def saveAsTable(self, table_name):
        Config().spark.data[table_name] = self.data

    def toPandas(self):
        return pd.DataFrame(data=self.data)


class SparkMock(Mock):

    data = {}

    @property
    def tables(self):
        return list(self.data.keys())

    def table(self, table_uri):

        df = DFMock(
            table_uri=table_uri,
            data=self.data[table_uri],
        )
        return df

    def sql(self, qry):

        if qry.startswith('SELECT 1 from'):
            for table in self.tables:
                if table in qry:
                    return True
            raise AnalysisException("Table does not exist")


def create_testing_config(
    working_dir,
    storage_key='',
    container_name='',
    account_name='',
    mock_spark=True,
    **kwargs
):
    """
    Creates a Config Object for testing purpose. Should be used when working with temporary
    directories.

    Args:
        working_dir (str): working dir for testing luisy tasks (mostly tmpdirs)
        storage_key (str): storage_key that can be set if no environ variable can be used in tests.
        account_name (str): Account name that can be set if no environ variable can be used in
            tests.
        container_name (str): Container name that can be set if no environ variable can be used in
            tests.
        **kwargs: params like `download` or `upload`
    """

    params = get_default_params(raw=False)
    params['working_dir'] = working_dir
    params['azure_storage_key'] = storage_key
    params['azure_container_name'] = container_name
    params['azure_account_name'] = account_name
    params['spark'] = SparkMock() if mock_spark else None

    for key, val in kwargs.items():
        params[key] = val
    Config().update(params)



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0


import luigi
import logging
import pandas as pd
import os
import traceback

from luisy.decorators import (
    pickle_output,
    auto_filename,
)
from luisy.config import Config
from luisy.visualize import visualize_task
from luisy.targets import CloudTarget

from luisy.helpers import RegexTaskPattern

logger = logging.getLogger(__name__)


@pickle_output
@auto_filename
class Task(luigi.Task):
    """
    Base task that provides interfaces for all decorators.
    """

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(name=self.__class__.__name__)

        # May  be set via decorator
        if not hasattr(self, 'project_name'):
            self.project_name = self.__class__.__module__.split('.')[0]

        if self.project_name == 'luisy':
            raise ValueError(
                'Something went wrong when setting the project name. '
                'Are you using the task correctly?'
            )

        super().__init__(*args, **kwargs)

    @classmethod
    def get_related_instances(cls, regex_placeholder='[\\S]*'):
        """
        Find executed instances of the task class on your local machine.

        Args:
            regex_placeholder (str): Regex placeholder that replaces value strings with a regex to
                find all matching instances on filesystem.

        Returns:
            list[luisy.Task]: list of task instances related to given class

        Raises:
            Exception: If problems in finding instances on the system occur, Exceptions are
                caught in general and the user is informed to look at his get_file_name() method.

        """
        try:
            regex_pattern = RegexTaskPattern(
                task_cls=cls,
                regex_placeholder=regex_placeholder
            )
            params = regex_pattern.get_available_params()
            return [cls(**entry) for entry in params.values()]
        except Exception:
            logger.info('Finding instances failed, have you overwritten Task.get_file_name() '
                        'method of your class?')
            logger.error(traceback.format_exc())
            raise

    def get_outdir(self):
        out_dir = Config().get_param('working_dir')
        out_dir = os.path.join(
            out_dir, self.get_sub_dir())

        return out_dir

    def get_file_name(self):
        """
        By overriding this method one can specify a custom filename.
        """
        pass

    def get_sub_dir(self):
        """
        Default implementation of a subdir.
        By overriding this method one can specify an additional subdir
        under the current working dir.
        """
        return ""

    def get_outfile(self):
        out_dir = self.get_outdir()
        os.makedirs(out_dir, exist_ok=True)
        file_name = self._remove_redundant_file_ending()
        return os.path.join(
            out_dir,
            f"{file_name}.{self.target_cls.file_ending}"
        )

    def _remove_redundant_file_ending(self):
        file_extension = "." + self.target_cls.file_ending
        if self.get_file_name().endswith(file_extension):
            return self.get_file_name()[:-len(file_extension)]
        else:
            return self.get_file_name()

    def download(self):
        """
        Downloads the output of the task from the cloud.
        """
        self.logger.info(f'Start downloading of {self}')
        self.output()._try_to_download()

    def upload(self, overwrite=False):
        """
        Uploads the output of the task to the cloud.abs

        Args:
            overwrite (bool): Whether an existing file in the cloud should be overwritten. Defaults
                to :code:`False`.
        """
        self.logger.info(f'Start uploading of {self}')
        self.output()._try_to_upload(overwrite=overwrite)

    def read_input(self):
        """Helper to read atomic inputs"""
        return self.input().read()

    def read(self):
        """
        Reads the output of the task and return is. If cloud synchronisation is activated, a file
        may be downloaded from the cloud
        """

        if not self.complete():
            raise ValueError('Task needs to be executed first, '
                             'try luisy.helpers.get_related_instances() to find executed '
                             'instances of this task')

        return self.output().read()

    def write(self, obj):
        target = self.output()
        target.make_dir(self.get_outdir())
        target.write(obj)

    def output(self):
        """
        """
        if issubclass(self.target_cls, CloudTarget):
            return self.target_cls(
                outdir=self.get_outdir(),
                **self.target_kwargs
            )
        else:
            return self.target_cls(
                path=self.get_outfile(),
                **self.target_kwargs)

    def clean(self):
        if isinstance(self, luigi.ExternalTask):
            raise ValueError('External tasks cannot be deleted that way')
        self.logger.info(f"Cleaning {self}")
        self.output().remove()

    def visualize(self, ax=None, unique_children=True, parameters_to_exclude=()):
        """
        Visualizes the dependencies tree of the task.

        Args:
            ax (matplotlib.axis.Axis): A matplotlib axis to draw in
            unique_children (bool): If true, add only subset of unique types of children
            parameters_to_exclude (list): List of names of py:mod:`luigi`-parameters to exclude in
                visualization.

        """
        visualize_task(
            task=self,
            ax=ax,
            unique_children=unique_children,
            parameters_to_exclude=parameters_to_exclude,
        )


class ExternalTask(Task, luigi.ExternalTask):
    pass


class ConcatenationTask(Task):
    def preprocess(self, df):
        return df

    def run(self):
        """
        """

        dfs = []
        for infile in self.input():
            self.logger.info('Read file {}'.format(infile.path))
            dfs.append(infile.read())

        df = pd.concat(dfs)
        df = self.preprocess(df)
        self.write(df)


class WrapperTask(Task, luigi.WrapperTask):
    def download(self):
        """
        Task cannot be downloaded, because no file exists.
        """
        pass

    def upload(self, overwrite=None):
        """
        Task cannot be uploaded, because no file exists.
        """
        pass

    def read(self):
        """
        Reads all the output of the required wrapped tasks. The type of the return value depends on
        the type of the return value of the :py:func:`WrapperTask.requires` method.
        """
        input_tasks = self.input()

        if isinstance(input_tasks, dict):
            return {
                k: task.read() for k, task in input_tasks.items()
            }

        if isinstance(input_tasks, list):
            return [task.read() for task in input_tasks]

        raise ValueError('Type of requires not understood')


class SparkTask(Task):
    @property
    def spark(self):
        return Config().spark

    def write(self, df):
        target = self.output()
        if target.requires_pandas and not isinstance(df, pd.DataFrame):
            df = df.toPandas()
        super().write(df)

    def read_input(self):
        df = self.input().read()
        if isinstance(df, pd.DataFrame):
            df = self.spark.createDataFrame(df)
        return df
